"""
Image transformations with coorx transforms carried along. 

Vision for the future: a collection of tools that can process images through a pipeline of transformations
and also output coorx transforms that map from one end of the pipeline to the other.

- rotate/scale around a specific center point
- get a cropped region that includes a specific point(s) with padding
- add padding to an image
- get an interpolated region
- given two partially overlapping images, give the region of overlap (and use this region to automatically slice from
  either image)
- resample / downscale with averaging


Basic use pattern:

    # make some transformations to an image
    img = Image(image_data)
    rotated = img.rotate(angle, center=(row, col))
    cropped = rotated[10:-10, 10:-10]

    # map coordinates from the original image to the cropped image
    pt2 = img.point([row, col]).mapped_to(cropped.system)

    # maybe add a physical coordinate system
    frame_tr = Transform(from_cs='physical', to_cs=img.system)
    pt3 = cropped.point([row, col]).mapped_to('physical')

    
Things to work out:
- how to handle 3d image stacks, rgb images, etc.
- how to handle slicing, indexing, fancy indexing    
- can we stack a bunch of transforms together and apply them all at once?
    tr = CompositeTransform([tr1, tr2, tr3])
    img2 = img.apply_transform(tr, new_shape)
- garbage collection of transforms and coordinate systems attached to images
    - ask graph to keep transforms/systems in weak references
"""
from __future__ import annotations

import numpy as np
import scipy

from .coordinates import Point, PointArray
from .linear import AffineTransform, TTransform, STTransform
from .systems import CoordinateSystemGraph, CoordinateSystem


class Image:
    """Wraps image data with a coordinate system and methods
    for transforming the image and mapping coordinates through the transforms.

    Parameters
    ----------
    image : ndarray
        The image data. Must be 2D or higher.
    axes : tuple, optional
        The axes of the image that correspond to spatial dimensions. Defaults to all axes.
    system : str | CoordinateSystem | None
        Optional name of the coordinate system to attach to the image.
    graph : str | CoordinateSystemGraph | None
        Optional graph to use for the coordinate system.
    """

    def __init__(self, image, axes=None, system=None, graph=None):
        if axes is None:
            axes = tuple(range(image.ndim))
        self.spatial_to_image_axes = np.asarray(axes)

        self.image = image
        self._parent_tr = None

        if graph is None:
            graph = system.graph if system is not None else 'image_graph'
        self.graph = CoordinateSystemGraph.get_graph(graph, create=True)

        if isinstance(system, CoordinateSystem):
            self.system = system
        else:
            if system is None:
                index = 0
                while True:
                    system = f'image_{index}'
                    if system not in self.graph.systems:
                        break
                    index += 1
            self.system = self.graph.add_system(system, ndim=self.spatial_ndim)

    @property
    def spatial_ndim(self):
        """Return the number of spatial dimensions of the image."""
        return len(self.spatial_to_image_axes)

    @property
    def spatial_shape(self):
        return tuple(self.image.shape[i] for i in self.spatial_to_image_axes)

    def point(self, coords):
        """Return a Point object with the given (row, col) coordinates in the CS of this image."""
        coords = np.asarray(coords)
        return Point(coords, system=self.system)

    def point_array(self, coords):
        """Return a PointArray object with the given (row, col) coordinates in the CS of this image."""
        coords = np.asarray(coords)
        return PointArray(coords, system=self.system)

    def rotate(self, angle, axes=(0, 1), **kwds):
        """Rotate the image by the given angle around the specified axes.

        Parameters
        ----------
        angle : float
            The angle in degrees to rotate the image.
        axes : (int, int), optional
            The two spatial axes involved in the rotation. Defaults to (0, 1). Beware: this is different from how
            AffineTransform's rotations work, and different from how Image.__init__ axes work.
        kwds : keyword arguments
            Additional keyword arguments to pass to `scipy.ndimage.rotate`.
        """
        img = self.image
        rotated_img = scipy.ndimage.rotate(img, angle, axes=self.spatial_to_image_axes[list(axes)], **kwds)
        img2 = self.copy(image=rotated_img)
        img2._parent_tr = self.make_rotation_transform(
            angle, axes, self.spatial_shape, img2.spatial_shape, from_cs=self.system, to_cs=img2.system
        )
        return img2

    def __getitem__(self, item):
        if not isinstance(item, tuple):
            item = (item,)
        if not all(isinstance(i, slice) for i in item):
            raise ValueError(
                f"Image.__getitem__ requires a tuple of slices, got {item}"
            )
        if len(item) != self.spatial_ndim:
            item = item + (slice(None),) * (self.spatial_ndim - len(item))
        cropped_img = self.image[item]
        img2 = self.copy(image=cropped_img)
        img2._parent_tr = self.make_crop_transform(
            item, self.image, from_cs=self.system, to_cs=img2.system
        )
        return img2

    def zoom(self, factors, **kwds):
        # fill in missing image axes with 1
        if np.isscalar(factors):
            factors = [factors] * self.spatial_ndim
        actual = np.ones(self.image.ndim, dtype=float)
        for i, ax in enumerate(self.spatial_to_image_axes):
            actual[ax] = factors[i]
        scaled_img = scipy.ndimage.zoom(self.image, actual, **kwds)

        img2 = self.copy(image=scaled_img)
        tr = STTransform(
            dims=(self.spatial_ndim, self.spatial_ndim),
            from_cs=self.system,
            to_cs=img2.system,
            cs_graph=self.graph,
        )
        tr.scale = factors
        img2._parent_tr = tr
        return img2

    def copy(self, **updates):
        kwds = {'axes': self.spatial_to_image_axes, 'graph': self.graph}
        if 'image' not in updates:
            kwds['image'] = self.image.copy()
        kwds.update(updates)
        return Image(**kwds)

    def make_rotation_transform(self, angle, axes, from_shape, to_shape, **kwds):
        # Make transform mapping unrotated to rotated coordinates
        from_center = np.array(from_shape) / 2
        to_center = np.array(to_shape) / 2
        tr = AffineTransform(dims=(self.spatial_ndim, self.spatial_ndim), cs_graph=self.graph, **kwds)
        tr.translate(-from_center)
        if self.spatial_ndim == 2:
            tr.rotate(-angle)
        elif self.spatial_ndim == 3:
            ax1 = np.zeros(self.spatial_ndim)
            ax1[axes[0]] = 1
            ax2 = np.zeros(self.spatial_ndim)
            ax2[axes[1]] = 1
            axis = np.cross(ax1, ax2)
            tr.rotate(-angle, axis=axis)
        else:
            raise ValueError("Image must be 2D or 3D for rotation")
        tr.translate(to_center)
        return tr

    def make_crop_transform(self, crop, img, **kwds):
        offset = [-crop[i].indices(img.shape[i])[0] for i in self.spatial_to_image_axes]
        return TTransform(offset=offset, dims=(self.spatial_ndim, self.spatial_ndim), cs_graph=self.graph, **kwds)



import numpy as np

from .base_transform import Transform
from .params import ArrayParameter, TransformParameter


class AxisSelectionEmbeddedTransform(Transform):
    """Wraps any transform in a larger-dimensional transform that passes specific axes to the wrapped
    transform, while keeping other axes unchanged.

    For example, you could use a 2D transform to affect just the x,y axes in a 3D space.
    """

    Linear = False
    Orthogonal = False
    NonScaling = False
    Isometric = False
    Equidimensional = False

    parameter_spec = [
        ArrayParameter("axes", dtype=int, default=lambda shape: np.arange(shape[0])),
        TransformParameter("transform"),
    ]

    def __init__(self, axes, transform, **kwds):
        super().__init__(**kwds, axes=axes, transform=transform)

    def _map(self, arr):
        out = arr.copy()
        out[:, self._state["axes"]] = self._state["transform"].map(arr[:, self._state["axes"]])
        return out

    def _imap(self, arr):
        out = arr.copy()
        out[:, self._state["axes"]] = self._state["transform"].imap(arr[:, self._state["axes"]])
        return out


class HomogeneousEmbeddedTransform(Transform):
    """Wraps any transform that uses homogeneous coordinates,
    allowing to operate with nonhomogeneous inputs/outputs instead.
    """

    Linear = False
    Orthogonal = False
    NonScaling = False
    Isometric = False
    Equidimensional = False

    parameter_spec = [
        TransformParameter("transform"),
    ]

    def __init__(self, transform, **kwds):
        expected_dims = (transform.dims[0] - 1, transform.dims[1] - 1)
        kwds.setdefault("dims", expected_dims)
        if kwds["dims"] != expected_dims:
            raise ValueError(
                f"Transform has dims {transform.dims}, expected "
                f"{(transform.dims[0]-1, transform.dims[1]-1)} for HomogeneousEmbeddedTransform"
            )
        super().__init__(**kwds, transform=transform)

    def _map(self, arr):
        out = self._state["transform"].map(self._to_homogeneous(arr))
        return self._from_homogeneous(out)

    def _imap(self, arr):
        out = self._state["transform"].imap(self._to_homogeneous(arr))
        return self._from_homogeneous(out)

    @staticmethod
    def _to_homogeneous(arr):
        hom = np.empty((arr.shape[0], arr.shape[1] + 1), dtype=arr.dtype)
        hom[:, :-1] = arr
        hom[:, -1] = 1
        return hom

    @staticmethod
    def _from_homogeneous(arr):
        return arr[:, :-1] / arr[:, -1:]



import warnings

import numpy as np

from .base_transform import Transform
from .params import ArrayParameter, TupleParameter


class LogTransform(Transform):
    """ND transform performing logarithmic transformation.

    Maps (x, y, z) => (log(base_x, x), log(base_y, y), log(base_z, z))

    Special base values:
    - None: Identity transformation (no change to that axis)
    - Negative values: Inverse (exponential) transformation (x => -base^x)
    - 1 or 0: Mathematically nonsensical but allowed (may produce inf/nan)
    - Fractional values: Standard logarithm (log(x)/log(base))

    Parameters
    ----------
    base : array-like
        Base values for each axis; length must be the same as the dimensionality of the transform.
        A base value of None provides identity transformation for that axis.
        Negative bases apply inverse (exponential) transformation.
    """

    Linear = False
    Orthogonal = True
    NonScaling = False
    Isometric = False
    Equidimensional = True

    parameter_spec = [
        TupleParameter("base", length="dims0", default=lambda shape: np.array([None] * shape[0])),
    ]

    def __init__(self, base=None, dims=None, **kwargs):
        super().__init__(dims, base=base, **kwargs)

    @property
    def base(self) -> tuple[float | None]:
        """
        *base* is a tuple containing the log base values that should be
        applied to each axis of the input vector. If any axis has a base == None,
        then that axis is not affected (identity transformation).
        """
        return self._state['base']

    @base.setter
    def base(self, s):
        self.set_params(base=s)

    def _map(self, coords):
        return self._map_with_base(coords, self._state['base'])

    def _imap(self, coords):
        return self._map_with_base(coords, [b if b is None else -b for b in self.base])

    @staticmethod
    def _map_with_base(coords, base):
        # Ensure output dtype can handle floating point values including NaN
        ret = _empty_array_like(coords)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)  # divide-by-zeros and invalid values
            for i in range(min(ret.shape[-1], 3)):
                if base[i] is None:
                    ret[..., i] = coords[..., i]
                elif base[i] > 0.0:
                    ret[..., i] = np.log(coords[..., i]) / np.log(base[i])
                else:  # base < 0 treated as inverse
                    ret[..., i] = (-base[i]) ** coords[..., i]

        ret[~np.isfinite(ret)] = np.nan  # set all non-finite values to NaN
        return ret

    def __repr__(self):
        return f"<LogTransform base={self._state['base']}>"


def _empty_array_like(data):
    output_dtype = data.dtype if data.dtype.kind == 'f' else np.float64
    return np.empty(data.shape, output_dtype)


class PolarTransform(Transform):
    """Polar transform

    Maps (theta, r, ...) to (x, y, ...), where `x = r*cos(theta)`
    and `y = r*sin(theta)`.
    """

    Linear = False
    Orthogonal = False
    NonScaling = False
    Isometric = False
    Equidimensional = True

    def __init__(self, dims=None, **kwargs):
        super().__init__(dims, **kwargs)

    def _map(self, coords):
        ret = _empty_array_like(coords)
        ret[..., 0] = coords[..., 1] * np.cos(coords[..., 0])
        ret[..., 1] = coords[..., 1] * np.sin(coords[..., 0])
        for i in range(2, coords.shape[-1]):  # copy any further axes
            ret[..., i] = coords[..., i]
        return ret

    def _imap(self, coords):
        ret = _empty_array_like(coords)
        ret[..., 0] = np.arctan2(
            coords[..., 1], coords[..., 0]
        )  # arctan2(y, x) for correct quadrant
        ret[..., 1] = (coords[..., 0] ** 2 + coords[..., 1] ** 2) ** 0.5
        for i in range(2, coords.shape[-1]):  # copy any further axes
            ret[..., i] = coords[..., i]
        return ret


# class SphericalTransform(Transform):
#    # TODO
#    pass


# class WarpTransform(Transform):
#    """ Multiple bilinear transforms in a grid arrangement.
#    """
#    # TODO


class LensDistortionTransform(Transform):
    """https://docs.opencv.org/4.x/dc/dbb/tutorial_py_calibration.html

    Coefficients are (k1, k2, p1, p2, k3)
    Where k1, k2, and k3 are radial distortion (coordinates are multiplied by 1 + k1*r^2 + k2*r^4 + k3*r^6),
    and p1, p2 are tangential distortion coefficients.
    """

    Linear = False
    Orthogonal = False
    NonScaling = False
    Isometric = False
    Equidimensional = True

    parameter_spec = [
        ArrayParameter(
            "coeff", dtype=float, shape=(5,), default=lambda shape: np.array([0, 0, 0, 0, 0])
        ),
    ]

    def __init__(self, **kwds):
        kwds.setdefault('dims', (2, 2))
        if kwds['dims'] != (2, 2):
            raise ValueError("LensDistortionTransform only supports 2D transforms")
        super().__init__(**kwds)

    @property
    def coeff(self):
        """
        *coeff* is an array of lens distortion coefficients (k1, k2, p1, p2, k3).
        """
        return self._state['coeff']

    def set_coeff(self, coeff):
        self.set_params(coeff=coeff)

    def _map(self, arr):
        k1, k2, p1, p2, k3 = self.coeff

        # radial distortion
        r = np.linalg.norm(arr, axis=1)
        dist = 1 + k1 * r**2 + k2 * r**4 + k3 * r**6
        out = arr * dist[:, None]

        # tangential distortion
        x = out[:, 0]
        y = out[:, 1]
        xy = x * y
        r2 = r**2
        out[:, 0] += 2 * p1 * xy + p2 * (r2 + 2 * x**2)
        out[:, 1] += 2 * p2 * xy + p1 * (r2 + 2 * y**2)

        return out

    def _imap(self, arr):
        """Inverse lens distortion mapping using iterative numerical method.

        Since lens distortion is nonlinear, we use Newton-Raphson iteration
        to find the undistorted coordinates that would map to the given distorted ones.
        """
        k1, k2, p1, p2, k3 = self.coeff

        # If all coefficients are zero, it's identity transform
        if all(c == 0 for c in self.coeff):
            return arr.copy()

        # Initial guess: use input as starting point
        undistorted = arr.copy()

        # Newton-Raphson iteration for inverse mapping
        for iteration in range(10):  # Maximum 10 iterations
            # Compute forward mapping of current guess
            forward = self._map(undistorted)

            # Compute residual (error)
            residual = forward - arr

            # Check convergence
            if np.allclose(residual, 0, atol=1e-8):
                break

            # Compute Jacobian matrix numerically
            eps = 1e-6
            jac = np.zeros((arr.shape[0], 2, 2))

            for i in range(2):
                perturbed = undistorted.copy()
                perturbed[:, i] += eps
                forward_perturbed = self._map(perturbed)
                jac[:, :, i] = (forward_perturbed - forward) / eps

            # Solve Jacobian * delta = -residual for delta
            delta = np.zeros_like(undistorted)
            try:
                # Use np.linalg.solve for each point
                for i in range(arr.shape[0]):
                    delta[i] = np.linalg.solve(jac[i], -residual[i])

                # Update guess
                undistorted += delta

            except np.linalg.LinAlgError:
                # If Jacobian is singular, use pseudo-inverse
                for i in range(arr.shape[0]):
                    delta[i] = np.linalg.pinv(jac[i]) @ (-residual[i])
                undistorted += delta

        return undistorted



import unittest
import inspect
import numpy as np

import coorx as coorx


def assert_eq(a, b):
    """Test for equivalence of deep data structures
    """
    assert type(a) == type(b)
    if isinstance(a, np.ndarray):
        assert a.shape == b.shape
        assert a.dtype == b.dtype
        assert np.allclose(a, b)
    elif isinstance(a, dict):
        assert len(a) == len(b)
        for k,v in a.items():
            assert k in b
            assert_eq(b[k], v)
    elif isinstance(a, (list, tuple)):
        assert len(a) == len(b)
        for x,y in zip(a, b):
            assert_eq(x, y)
    else:
        assert a == b

    
class InitTests(unittest.TestCase):
    def test_transform_types(self):
        types = list(set(
                    self.list_types(coorx.base_transform) +
                    self.list_types(coorx.linear) + 
                    self.list_types(coorx.nonlinear) +
                    self.list_types(coorx.composite) +
                    self.list_types(coorx.util)
                ))
        types.remove(coorx.Transform)
        
        found_types = coorx.transform_types()
        for typ in types:
            assert typ in found_types
            found_types.remove(typ)
        assert len(found_types) == 0

        class TestTransform(coorx.Transform):
            pass
        
        assert TestTransform in coorx.transform_types()
                    
    def list_types(self, mod):
        return [x for x in mod.__dict__.values() if inspect.isclass(x) and issubclass(x, coorx.Transform)]

    def test_create(self):
        for typ in coorx.transform_types():
            if typ in [coorx.InverseTransform, coorx.CompositeTransform, coorx.SimplifiedCompositeTransform]:
                continue

            kwargs = {
                coorx.nonlinear.LogTransform: {'base': [2, 3, 5], 'dims': (3, 3)},
                coorx.nonlinear.LensDistortionTransform: {'coeff': (0.1, 0.2, 0.3, 0.4, 0.5), 'dims': (2, 2)},
                coorx.util.AxisSelectionEmbeddedTransform: {'dims': (3, 3), 'axes': [0, 1], 'transform': coorx.NullTransform(dims=(2, 2))},
                coorx.util.HomogeneousEmbeddedTransform: {'dims': (3, 3), 'transform': coorx.NullTransform(dims=(4, 4))},
                coorx.linear.BilinearTransform: {'dims': (2, 2)},
                coorx.linear.TransposeTransform: {'axis_order': [1, 0, 2]},
                coorx.linear.Homography2DTransform: {'dims': (2, 2)},
            }
            default_kwargs = {'dims': (3, 3)}
            transform1 = typ(**kwargs.get(typ, default_kwargs))
            state = transform1.save_state()
            transform2 = coorx.create_transform(**state)
            assert type(transform1) is type(transform2)
            assert_eq(state, transform2.save_state())



from __future__ import annotations

from ._types import CoordSysOrStr, GraphOrGraphName


def __reload__(old):
    # when this module is reloaded, we want to preserve existing graphs
    CoordinateSystemGraph.all_graphs = old["CoordinateSystemGraph"].all_graphs


def get_coordinate_system(system: CoordSysOrStr, graph: GraphOrGraphName=None, ndim=None, create=False) -> 'CoordinateSystem':
    """Return a CoordinateSystem instance.
    """
    if graph is None:
        if isinstance(system, CoordinateSystem):
            graph = system.graph
        else:
            graph = 'default'
    csg = CoordinateSystemGraph.get_graph(graph)
    return csg.check_system(system, ndim=ndim, create=create)


class CoordinateSystemGraph:
    """Multiple coordinate systems connected by transforms.

    A CoordinateSystemGraph keeps track of the relationships between many coordinate systems.
    This makes it possible to automatically determine a chain of transforms that map
    between any two coordinate systems (as long as they are indirectly connected within
    the graph).

    CoordinateSystemGraph also provides basic sanity checking:
    - Only one coordinate system of each name, and that it has the expected properties
      (such as dimensionality)
    - Only one transform connecting any pair of coordinate systems
      (requires unique_transforms=True; the default graph ignores this constraint)

    Caveats
    =======
    * setting a CoordinateSystemGraph to have unique_transforms is incompatible with
      using `as_affine` or anything that depends on that (`full_matrix`, `as_vispy`,
      etc.)
    """

    all_graphs: dict[str|None, 'CoordinateSystemGraph'] = {}

    @classmethod
    def get_graph(cls, graph_name=None, create=False, **create_kwargs) -> 'CoordinateSystemGraph':
        if isinstance(graph_name, CoordinateSystemGraph):
            return graph_name
        if graph_name is None:
            graph_name = 'default'        
        if graph_name not in cls.all_graphs:
            if not create:
                raise NameError(f"No coordinate system graph named '{graph_name}' exists")
            defaults = {'unique_transforms': False}
            defaults.update(create_kwargs)
            graph = CoordinateSystemGraph(name=graph_name, **defaults)
            cls.all_graphs[graph_name] = graph
        else:
            graph = cls.all_graphs[graph_name]
            create_kwargs = create_kwargs.copy()
            unique_transforms = create_kwargs.pop('unique_transforms', None)
            if unique_transforms is not None:
                assert graph.unique_transforms == unique_transforms, "Cannot change unique_transforms setting of existing graph"
            if create_kwargs:
                raise TypeError(f"get_graph got unexpected keyword arguments: {list(create_kwargs.keys())}")
        return graph

    def __init__(self, name: str|None, unique_transforms=False):
        assert name not in self.all_graphs, f"A coordinate system graph named {name} already exists."
        self.all_graphs[name] = self
        self.name = name
        self.unique_transforms = unique_transforms
        self.systems = {}  # maps {system_name: CoordinateSystem}
        self.transforms = {}  # maps {cs1: {cs2: transform, ...}, ...}

    def __repr__(self):
        return f"<CoordinateSystemGraph {self.name}>"
    
    def __str__(self):
        return str(self.name)

    def add_system(self, name: str, ndim: int):
        assert name not in self.systems, f"A system named '{name}' is already added to this graph"
        assert isinstance(ndim, int)
        sys = CoordinateSystem(name=name, ndim=ndim, graph=self)
        self.systems[name] = sys
        return sys

    def get_system(self, name):
        return self.systems[name]

    def add_transform(self, transform: "Transform", from_cs: CoordSysOrStr, to_cs: CoordSysOrStr):
        # look up coordinate systems
        cs = (
            self.check_system(from_cs, ndim=transform.dims[0], create=True),
            self.check_system(to_cs, ndim=transform.dims[1], create=True),
        )

        # make sure no transform exists linking these systems
        if self.unique_transforms:
            have_transform_already = cs[0] in self.transforms and cs[1] in self.transforms[cs[0]]
            have_inverse_already = cs[1] in self.transforms and cs[0] in self.transforms[cs[1]]
            assert not have_transform_already, f"A transform is already added connecting '{from_cs}' to '{to_cs}'"
            assert not have_inverse_already, f"A transform is already added connecting '{to_cs}' to '{from_cs}'"

        # make sure this transform has not been assigned elsewhere
        assert transform._systems == (
            None,
            None,
        ), "This transform already connects a different set of coordinate systems"
        transform._systems = cs

        # record the new transform connection
        self.transforms.setdefault(cs[0], {})[cs[1]] = transform

    def check_system(self, system: CoordSysOrStr, ndim: int|None=None, create: bool = False):
        """Check that a system exists with the given name and ndim.

        If `create` is True, then the system may be created if it does not already exist.

        Return the named CoordinateSystem instance.
        """
        # get the CoordinateSystem instance, creating if needed
        if isinstance(system, str) and system not in self.systems and create:
            assert isinstance(ndim, int), "Must specify ndim when creating a new system"
            cs = self.add_system(system, ndim)
        else:
            cs = self.system(system)

        # check ndim is correct
        if ndim is not None and cs.ndim != ndim:
            raise TypeError(f"System '{cs}' is {cs.ndim}D (expected {ndim}D)")

        return cs

    def system(self, system: CoordSysOrStr) -> "CoordinateSystem":
        """Return a CoordinateSystem belonging to this graph."""
        if isinstance(system, str):
            if system not in self.systems:
                raise NameError(f"No coordinate system named '{system}' in this graph")
            return self.systems[system]
        elif isinstance(system, CoordinateSystem):
            assert system.graph is self, f"CoordinateSystem {system} belongs to graph {system.graph} (expected {self})"
            return system
        else:
            raise TypeError("system must be str or CoordinateSystem instance")

    def transform(self, from_cs: CoordSysOrStr, to_cs: CoordSysOrStr) -> "Transform":
        """Return the transform linking from_cs to to_cs, or raise TypeError if no linking path exists."""
        try:
            return self._direct_transform(from_cs, to_cs)
        except TypeError:
            # if no direct transform exists, try to find a path
            return self.transform_chain(self.transform_path(from_cs, to_cs))

    def _direct_transform(self, from_cs: CoordSysOrStr, to_cs: CoordSysOrStr) -> "Transform":
        """Return the transform linking cs1 to cs2, or raise KeyError if none is defined.
        """
        # check that coordinate systems exist
        from_cs, to_cs = self.system(from_cs), self.system(to_cs)

        fwd = self.transforms.get(from_cs, {}).get(to_cs, None)
        if fwd is not None:
            return fwd
        inv = self.transforms.get(to_cs, {}).get(from_cs, None)
        if inv is not None:
            return inv.inverse

        raise TypeError(f"No transform defined linking '{from_cs}' to '{to_cs}'")

    def transform_path(self, start, end) -> list:
        """Return a list of transforms needed to map from start to end."""
        start, end = self.system(start), self.system(end)
        if start == end:
            return []
        if found := self._find_path(start, end, set()):
            return [start] + found
        raise TypeError(f"No transform path from {start} to {end}")

    def _find_path(self, start, end, visited) -> list | None:
        if start in visited:
            return None
        visited.add(start)
        from_start = self.transforms.get(start, {})
        if end in from_start:
            return [end]
        for next_cs in from_start:
            path = self._find_path(next_cs, end, visited)
            if path is not None:
                return [next_cs] + path
        to_start = {cs for cs in self.transforms if start in self.transforms[cs]}
        if end in to_start:
            return [end]
        for next_cs in to_start:
            path = self._find_path(next_cs, end, visited)
            if path is not None:
                return [next_cs] + path
        return None

    def transform_chain(self, systems) -> 'CompositeTransform':
        from .composite import CompositeTransform
        transforms = []
        for i in range(1, len(systems)):
            cs1, cs2 = systems[i - 1 : i + 1]
            transforms.append(self._direct_transform(cs1, cs2))
        return CompositeTransform(transforms)


class CoordinateSystem:
    def __init__(self, name: str, ndim: int, graph: CoordinateSystemGraph):
        self.graph = graph
        self.name = name
        self.ndim = ndim

    def __repr__(self):
        if self.graph.name is not None:
            name = f"{self.graph.name}:{self.name}"
        else:
            name = self.name
        return f"<CoordinateSystem {name}[{self.ndim}]>"

    def __str__(self):
        return self.name

    def save_state(self):
        return {"type": type(self).__name__, "name": self.name, "ndim": self.ndim, "graph": self.graph.name}

    def get_transform_to(self, to_cs: CoordSysOrStr) -> "Transform":
        """Return the transform linking this coordinate system to another."""
        return self.graph.transform(self, to_cs)


default_cs_graph = CoordinateSystemGraph(name='default', unique_transforms=False)



from typing import Union, Protocol

import numpy as np

StrOrNone = Union[None, str]
Dim = Union[None, int]
Dims = Union[Dim, 'tuple[Dim, Dim]']
CoordSysOrStr = Union[str, 'CoordinateSystem']
GraphOrGraphName = Union[str, None, 'CoordinateSystemGraph']


class CustomMappable(Protocol):
    def _coorx_transform(self, tr):
        pass


Mappable = Union[np.ndarray, list, tuple, CustomMappable]



import contextlib
from copy import deepcopy
from typing import Callable

import numpy as np

from ._types import Dims, StrOrNone, Mappable
from .events import CallbackRegistry, ChangeEvent
from .params import Parameter, TransformParameter
from .systems import CoordinateSystemGraph, CoordinateSystem


class DependentTransformError(Exception):
    """Raised when an operation is not allowed on certain dependent transforms."""

    pass


class Transform(object):
    """
    Transform is a base class that defines a pair of complementary
    coordinate mapping functions in both python and GLSL.

    All Transform subclasses define map() and imap() methods that map
    an object through the forward or inverse transformation, respectively.

    Optionally, an inverse() method returns a new transform performing the
    inverse mapping.

    Note that although all classes should define both map() and imap(), it
    is not necessarily the case that imap(map(x)) == x; there may be instances
    where the inverse mapping is ambiguous or otherwise meaningless.

    """

    # Flags used to describe the transformation. Subclasses should define each
    # as True or False.
    # (usually used for making optimization decisions)

    # If True, then for any set of colinear points, the
    # transformed points will also be colinear.
    Linear = None

    # The transformation's effect on one axis is independent
    # of the input position along any other axis.
    Orthogonal = None

    # If True, then the distance between two input points is the
    # same as the distance between the transformed points.
    NonScaling = None

    # Scale factors are applied equally to all axes.
    Isometric = None

    # Multiplying the input by a scale factor causes the output to be multiplied by
    # the same factor:  Tr(s * x) = s * Tr(x)
    Homogeneous = None

    # The transform of two added vectors is the same as the sum of the individually
    # transformed vectors:  T(a + b) = T(a) + T(b)
    Additive = None

    # If True, then input and output dimensionality must be the same.
    Equidimensional = None

    # Specification of parameters for this transform. Each subclass should
    # define its own parameter_spec list.
    parameter_spec: list[Parameter] = []

    def __init__(
        self,
        dims: Dims = None,
        dynamic: bool = False,
        from_cs: StrOrNone = None,
        to_cs: StrOrNone = None,
        cs_graph: StrOrNone = None,
        **kwargs,
    ):
        self._state = {"dynamic": dynamic, "dims": self._validate_dims(dims, **kwargs)}
        self._init_with_no_state()

        # optional coordinate system tracking
        if from_cs is not None:
            self.set_systems(from_cs, to_cs, cs_graph)
        self.set_params(**kwargs)

    def _init_with_no_state(self):
        self._change_callbacks = CallbackRegistry()
        self._inverse = None
        self._systems = (None, None)

    def _validate_dims(self, dims: None | int | tuple[int, int] | list[int], **kwargs):
        """Determine dimensionality from parameters or *dims* argument.

        If *dims* has a value, then it must be a tuple (in, out) or int and it must agree with the
        length of all provided parameters. If *dims* is not provided, then it is determined from the
        length of the parameters that use dims, which must all be in agreement.
        """
        if dims is not None:
            # make sure dims agrees with all parameters that use dims
            if isinstance(dims, int):
                dims = (dims, dims)
            # todo yaml loads tuples as lists; accept that too?
            if not isinstance(dims, (tuple, list)) or len(dims) != 2:
                raise ValueError(f"dims must be an int or a tuple/list of two ints, not {dims}")
            for spec in self.parameter_spec:
                if spec.name in kwargs:
                    spec.check_dims(kwargs[spec.name], dims)
        else:
            # infer dims from parameters
            dims = [None, None]
            for spec in self.parameter_spec:
                if spec.name not in kwargs:
                    continue
                this_param_dims = spec.infer_dims(kwargs[spec.name])
                for i, dim in enumerate(this_param_dims):
                    if dim is not None:
                        if dims[i] is None:
                            dims[i] = dim
                        elif dims[i] != dim:
                            raise ValueError(
                                f"Could not determine dimensionality of transform: parameter {spec.name}"
                                f"({this_param_dims}) does not match previously inferred dims ({dims})"
                            )

            if 0 in dims:
                raise ValueError("Transform dimensionality cannot be zero")
            if self.Equidimensional and None in dims:
                # enforce equidimensionality
                dims = (dims[0] or dims[1], dims[1] or dims[0])
            # check that dims are now fully specified
            if dims[0] is None or dims[1] is None:
                if dims[0] is None and dims[1] is None:
                    extra = ""
                elif dims[0] is None:
                    extra = "input "
                else:
                    extra = "output "
                raise ValueError(
                    f"Could not determine {extra}dimensionality of transform; specify dims or "
                    "parameters that imply dims."
                )

        if self.Equidimensional and dims[0] != dims[1]:
            raise ValueError("Equidimensional transforms must have equal input and output dims")
        return tuple(dims)

    @property
    def dims(self):
        """Tuple holding the (input, output) dimensions for this transform."""
        return self._state['dims']

    @property
    def dynamic(self):
        """Boolean flag that indicates whether this transform is expected to
        change frequently.

        Transforms that are flagged as dynamic will not be collapsed in
        ``CompositeTransform.simplified``. This allows changes to the transform
        to propagate through the chain without requiring the chain to be
        re-simplified.
        """
        return self._state['dynamic']

    @dynamic.setter
    def dynamic(self, d):
        self._state['dynamic'] = d  # no spec, so not through set_params

    @property
    def systems(self) -> tuple[CoordinateSystem | None, CoordinateSystem | None]:
        """The CoordinateSystem instances mapped from and to by this transform."""
        return self._systems

    def set_systems(self, from_cs, to_cs, cs_graph=None):
        assert (from_cs is None) == (
            to_cs is None
        ), "from_cs and to_cs must both be None or both be coordinate systems"
        if from_cs is not None:
            if cs_graph is None and isinstance(from_cs, CoordinateSystem):
                cs_graph = from_cs.graph
            else:
                cs_graph = CoordinateSystemGraph.get_graph(cs_graph)
            cs_graph.add_transform(self, from_cs=from_cs, to_cs=to_cs)

    def map(self, obj: Mappable):
        """
        Return *obj* mapped through the forward transformation.

        Parameters
        ----------
            obj : tuple (x,y) or (x,y,z)
                  array with shape (..., 2) or (..., 3)
        """
        return self._prepare_and_map(obj)

    def _map(self, arr):
        """Map a 2D array (n_pts, n_dims) through this transform.

        This method must be redefined in sublcasses.
        """
        raise NotImplementedError

    def imap(self, obj: Mappable):
        """
        Return *obj* mapped through the inverse transformation.

        Parameters
        ----------
            obj : tuple (x,y) or (x,y,z)
                  array with shape (..., 2) or (..., 3)
        """
        return self.inverse.map(obj)

    def _imap(self, arr):
        """Map a 2D array (n_pts, n_dims) through the inverse of this transform.

        This method may be redefined in sublcasses.
        """
        raise NotImplementedError

    def _prepare_and_map(self, obj: Mappable):
        """
        Convert a mappable object to a 2D numpy array, pass it through this Transform's _map method,
        then convert and return the result.

        The Transform's _map method will be called with a 2D array
        of shape (N, M), where N is the number of points and M is the number of dimensions.
        Accepts lists, tuples, and arrays of any dimensionality and flattens extra dimensions into N.
        After mapping, any flattened axes are re-expanded to match the original input shape.

        For list, tuple, and array inputs, the return value is a numpy array of the same shape as
        the input, with the exception that the last dimension is determined only by the return value.

        Alternatively, any class may determine how to map itself by defining a _coorx_transform()
        method that accepts this transform as an argument.
        """
        if hasattr(obj, '_coorx_transform'):
            # let the object decide how to apply this transform
            return obj._coorx_transform(tr=self)
        elif isinstance(obj, (tuple, list, np.ndarray)):
            arr_2d, original_shape = self._prepare_arg_for_mapping(obj)
            if self.dims[0] not in (None, arr_2d.shape[1]):
                raise TypeError(
                    f"Transform maps from {self.dims[0]}D, but data to be mapped is {arr_2d.shape[1]}D"
                )
            ret = self._map(arr_2d)
            assert ret.ndim == 2
            assert self.dims[1] in (
                None,
                ret.shape[1],
            ), f"Transform maps to {self.dims[1]}D, but mapping generated {ret.shape[1]}D"
            return self._restore_shape(ret, original_shape)
        elif hasattr(obj, '__len__') and len(obj) == self.dims[0]:
            try:
                # fudge for single points passed as list/tuple-like objects
                arr = np.asarray(obj)
                ret = self._map(arr)
                return type(obj)(*ret)
            except Exception as e:
                raise TypeError(f"Cannot directly map object through Transforms: {obj}") from e
        else:
            raise TypeError(f"Cannot use argument for mapping: {obj}")

    @staticmethod
    def _prepare_arg_for_mapping(arg):
        """Convert arg to a 2D numpy array.

        If the argument ndim is > 2, then all dimensions except the last are flattened.

        Return the reshaped array and a tuple containing the original shape.
        """
        arg = np.asarray(arg)
        original_shape = arg.shape
        arg = arg.reshape(int(np.prod(arg.shape[:-1])), arg.shape[-1])
        return arg, original_shape

    @staticmethod
    def _restore_shape(arg, shape):
        """Return an array with shape determined by shape[:-1] + (arg.shape[-1],)"""
        if arg is None:
            return arg
        return arg.reshape(shape[:-1] + (arg.shape[-1],))

    @property
    def inverse(self):
        """The inverse of this transform."""
        if self._inverse is None:
            self._inverse = InverseTransform(self)
        return self._inverse

    @property
    def params(self):
        """Return a dict of parameters specifying this transform."""
        return deepcopy(self._state)

    def set_params(self, **kwds):
        """Set parameters specifying this transform.

        Parameter names must be the same as the keys in self._state.
        """
        any_changed = False
        for name in kwds:
            validator = self.get_validator(name)
            value, this_changed = validator.validate(kwds[name], self._state.get(name), self.dims)

            if this_changed:
                any_changed = True
                self._state[name] = value
        if any_changed:
            self._update()

    @classmethod
    def get_validator(cls, param_name: str):
        if param_name not in cls.param_spec_dict():
            raise NameError(f"Transform {cls.__name__} has no parameter '{param_name}'")
        return cls.param_spec_dict()[param_name]

    @classmethod
    def param_spec_dict(cls):
        """Return a dict mapping parameter names to Parameter instances for this class."""
        if not hasattr(cls, '_param_spec_dict'):
            cls._param_spec_dict = {p.name: p for p in cls.parameter_spec}
        return cls._param_spec_dict

    def as_affine(self):
        """Return an equivalent affine transform if possible."""
        raise NotImplementedError()

    @property
    def full_matrix(self) -> np.ndarray:
        """
        Return the full transformation matrix for this transform, if possible.

        Modifying the returned array has no effect on the transform instance that generated it.
        """
        return self.as_affine().full_matrix

    def as_vispy(self):
        """Return a VisPy transform that is equivalent to this transform, if possible."""
        if self.dims != (3, 3):
            raise NotImplementedError("as_vispy is only implemented for 3D transforms")
        from vispy.visuals.transforms import MatrixTransform

        # a functional default if nothing else is implemented
        return MatrixTransform(self.full_matrix.T)

    def as_pyqtgraph(self):
        """Return a PyQtGraph transform that is equivalent to this transform, if possible."""
        from pyqtgraph import SRTTransform3D
        from pyqtgraph.Qt import QtGui

        # a functional default if nothing else is implemented
        return SRTTransform3D(QtGui.QMatrix4x4(self.full_matrix.reshape(-1)))

    def to_qmatrix4x4(self, QtGui=None):
        """Return a QMatrix4x4 that is equivalent to this transform."""
        from .qt import import_qt_gui

        return import_qt_gui().QMatrix4x4(self.full_matrix.reshape(-1))

    def add_change_callback(self, cb: Callable[[ChangeEvent], None], keep_reference: bool = False, duplicates: str = 'error'):
        """Add a callback that will be called whenever parameters of this transform change. If
        keep_reference is False, the callback will be held with a weak reference. This allows the
        object and its context to be garbage collected. Typically, keep_reference should be False
        whenever using a bound method, and True when using a standalone function whose side effects
        are desired even if no other references to that function exist.

        Parameters
        ----------
        cb : Callable[[ChangeEvent], None]
            The callback to register. It must accept a single argument, which will be a ChangeEvent
            instance describing the change.
        keep_reference : bool, optional
            Whether to keep a strong reference to the callback, by default False.
        duplicates : {'error', 'ignore', 'add'}, optional
            How to handle duplicate registrations of the same callback.
            'error' raises a ValueError, 'ignore' silently avoids adding duplicate callbacks
            (although it will upgrade a weakly-referenced callback to strongly-referenced, if requested), 
            'add' registers the callback multiple times. Default is 'error'.
        """
        self._change_callbacks.add(cb, keep_reference=keep_reference, duplicates=duplicates)

    def remove_change_callback(self, cb):
        """Remove a change callback."""
        self._change_callbacks.remove(cb)

    def _update(self, source_event=None):
        """
        Called to inform any listeners that this transform has changed.
        """
        event = ChangeEvent(transform=self, source_event=source_event)
        self._change_callbacks(event)

    def __mul__(self, tr):
        """
        Transform multiplication returns a new transform that is equivalent to
        the two operands performed in series.

        By default, multiplying two Transforms `A * B` will return
        ChainTransform([A, B]). Subclasses may redefine this operation to
        return more optimized results.

        To ensure that both operands have a chance to simplify the operation,
        all subclasses should follow the same procedure. For `A * B`:

        0. The inner systems of A and B must be the same (all
           implementations must call validate_transform_for_mul to ensure this).
        1. A.__mul__(B) attempts to generate an optimized transform product.
        2. If that fails, it must:

               * return super(A).__mul__(B) OR
               * return NotImplemented if the superclass would return an
                 invalid result.

        3. When Transform.__mul__(A, B) is called, it returns
           NotImplemented, which causes B.__rmul__(A) to be invoked.
        4. B.__rmul__(A) attempts to generate an optimized transform product.
        5. If that fails, it must:

               * return super(B).__rmul__(A) OR
               * return ChainTransform([B, A]) if the superclass would return
                 an invalid result.

        6. When Transform.__rmul__(B, A) is called, ChainTransform([A, B])
           is returned.
        """
        # switch to __rmul__ attempts.
        # Don't use the "return NotImplemented" trick, because that won't work if
        # self and tr are of the same type.
        return tr.__rmul__(self)

    def __rmul__(self, tr):
        """tr * self"""
        tr.validate_transform_for_mul(self)
        return CompositeTransform([self, tr])

    def __repr__(self):
        return f"<{self.__class__.__name__} at 0x{id(self):x}>"

    def __getstate__(self):
        state = self._state.copy()
        if self.systems[0] is None:
            state['systems'] = (None, None, None)
        else:
            state['systems'] = (
                self.systems[0].name,
                self.systems[1].name,
                self.systems[0].graph.name,
            )
        return state

    def __setstate__(self, state):
        self._init_with_no_state()
        from_cs, to_cs, *graph = state.pop('systems', (None, None, None))
        if len(graph) == 0:
            graph = None
        else:
            graph = graph[0]

        self._state = {k: v for k, v in state.items() if k not in self.param_spec_dict()}
        params = {k: v for k, v in state.items() if k in self.param_spec_dict()}
        self._state["dims"] = self._validate_dims(self._state.get("dims", None), **params)
        self.set_params(**params)
        with contextlib.suppress(DependentTransformError):
            self._systems = (None, None)
            self.set_systems(from_cs, to_cs, graph)

    def copy(self, from_cs=None, to_cs=None):
        """Return a copy of this transform."""
        state = self.save_state()
        if from_cs is not None or to_cs is not None:
            from_cs = from_cs or self.systems[0]
            to_cs = to_cs or self.systems[1]
            graph = None
            if from_cs is not None and not isinstance(from_cs, str):
                graph = from_cs.graph.name
            if graph is None and to_cs is not None and not isinstance(to_cs, str):
                graph = to_cs.graph.name
            state['systems'] = (from_cs, to_cs)
            state['graph'] = graph
        return self.from_state(state)

    def save_state(self):
        """Return serializable parameters that specify this transform. Distinct from __getstate__
        for no good long-term reason, but we need to support yaml serialization somehow for now."""

        def to_simple(o):
            if isinstance(o, Transform):
                return o.save_state()
            elif np.isscalar(o) or o is None:
                return o
            else:
                return [to_simple(e) for e in np.asarray(o).tolist()]

        return {
            'type': type(self).__name__,
            **{k: to_simple(v) for k, v in self.__getstate__().items()},
        }

    @classmethod
    def from_state(cls, state):
        """Return a Transform instance created from saved state."""
        tr = cls.__new__(cls)
        tr.__setstate__(state)
        return tr

    def __eq__(self, tr):
        if type(self) is not type(tr):
            return False
        if self._state.keys() != tr._state.keys():
            return False
        for k in self._state:
            v1 = self._state[k]
            v2 = tr._state[k]
            if np.isscalar(v1):
                if not (np.isscalar(v2) and v1 == v2):
                    return False
            else:
                if not np.all(np.asarray(v1) == np.asarray(v2)):
                    return False
        return True

    def validate_transform_for_mul(self, tr):
        if tr.systems[1] != self.systems[0]:
            raise TypeError(
                f"Cannot multiply transforms with different inner coordinate systems: {self.systems[0]} != {tr.systems[1]}"
            )


class InverseTransform(Transform):
    parameter_spec = [
        TransformParameter("inverse"),
    ]

    def __init__(self, transform):
        super().__init__(inverse=transform)

    def _validate_dims(self, dims, **kwargs):
        # dims are determined by the inverse transform
        return None

    def set_systems(self, from_cs, to_cs, cs_graph=None):
        raise DependentTransformError("Cannot set systems on a dependent inverse transform")

    def as_affine(self):
        affine = self._state["inverse"].as_affine()
        return type(affine)(
            matrix=affine.inv_matrix,
            offset=affine.inv_matrix @ affine.inv_offset,
            from_cs=self.systems[0],
            to_cs=self.systems[1],
        )

    def copy(self, from_cs=None, to_cs=None):
        return self._state['inverse'].copy(from_cs=to_cs, to_cs=from_cs).inverse

    def set_params(self, inverse):
        if isinstance(inverse, Transform):
            self._state['inverse'] = inverse
        else:
            from . import create_transform

            self._state['inverse'] = create_transform(**inverse)
        self._map = self._state['inverse']._imap
        self._imap = self._state['inverse']._map
        self._update()

    @property
    def inverse(self):
        return self._state['inverse']

    @property
    def dims(self):
        return self._state['inverse'].dims[::-1]

    @property
    def systems(self):
        return self._state['inverse'].systems[::-1]

    @property
    def Linear(self):
        return self._state['inverse'].Linear

    @property
    def Orthogonal(self):
        return self._state['inverse'].Orthogonal

    @property
    def NonScaling(self):
        return self._state['inverse'].NonScaling

    @property
    def Isometric(self):
        return self._state['inverse'].Isometric

    def __repr__(self):
        return f"<Inverse of {self._state['inverse']!r}>"


# import here to avoid import cycle; needed for Transform.__mul__.
from .composite import CompositeTransform



import contextlib
import importlib
import sys

HAVE_QT = None
QtGui = None
qt_lib_order = ['PyQt6', 'PySide6', 'PyQt5', 'PySide2']


def check_qt_imported():
    global HAVE_QT, QtGui
    if HAVE_QT is None:
        for qt_lib in qt_lib_order:
            if f'{qt_lib}.QtGui' in sys.modules:
                HAVE_QT = True
                QtGui = importlib.import_module(f'{qt_lib}.QtGui')
                break
    return HAVE_QT


check_qt_imported()


def import_qt_gui():
    """Import QtGui and return it."""
    global HAVE_QT, QtGui
    if not HAVE_QT:
        check_qt_imported()

    if QtGui is not None:
        return QtGui

    if not HAVE_QT:
        HAVE_QT = False
        for qt_lib in qt_lib_order:
            with contextlib.suppress(ImportError):
                QtGui = importlib.import_module(f'{qt_lib}.QtGui')
                HAVE_QT = True
                return QtGui

    raise ImportError(f'No importable Qt library found (tried {", ".join(qt_lib_order)})')



"""
Library of matrix transformation functions adapted from vispy.
"""

from __future__ import division

# Note: we use functions (e.g. sin) from math module because they're faster

import math
import numpy as np
import numpy.linalg


def translate(offset, dtype=None):
    """Return an N-dimensional translation matrix (of shape N+1,N+1)

    Parameters
    ----------
    offset : array-like
        Amount to translate each axis
    dtype : dtype | None
        Output type (if None, then same as *offset*).

    Returns
    -------
    M : ndarray
        Transformation matrix describing the translation.
    """
    if dtype is None:
        dtype = offset.dtype
    M = np.eye(len(offset) + 1, dtype=dtype)
    M[-1, :len(offset)] = offset
    return M


def scale(s, dtype=None):
    """Return an N-dimensional scaling matrix (of shape N+1,N+1)

    Parameters
    ----------
    scale : array-like, shape (3,)
        Scale factors for each axis
    dtype : dtype | None
        Output type (if None, then same as *scale*).

    Returns
    -------
    M : ndarray
        Transformation matrix describing the scaling.
    """
    if dtype is None:
        dtype = scale.dtype
    return np.array(np.diag(np.concatenate([s, (1.,)])), dtype)


def rotate2d(angle, dtype=None):
    """Return a 2D rotation matrix (shape 2,2)

    Parameters
    ----------
    angle : float
        The angle of rotation, in degrees.
    dtype : dtype | None
        Output dtype
    """
    rad = np.radians(angle)
    c, s = math.cos(rad), math.sin(rad)
    M = np.array([
        [c,  s],
        [-s, c],
    ], dtype=dtype)
    return M


def rotate3d(angle, axis, dtype=None):
    """Return a 3D rotation matrix (shape 3,3) for rotation about a vector.

    Parameters
    ----------
    angle : float
        The angle of rotation, in degrees.
    axis : ndarray
        The x, y, z coordinates of the axis direction vector.
    dtype : dtype | None
        Output dtype
    """
    angle = np.radians(angle)
    assert len(axis) == 3
    x, y, z = axis / np.linalg.norm(axis)
    c, s = math.cos(angle), math.sin(angle)
    cx, cy, cz = (1 - c) * x, (1 - c) * y, (1 - c) * z
    M = np.array([
        [cx * x + c, cy * x - z * s, cz * x + y * s],
        [cx * y + z * s, cy * y + c, cz * y - x * s],
        [cx * z - y * s, cy * z + x * s, cz * z + c],
    ], dtype).T
    return M


def affine_map(points1, points2):
    """ Find an N-D affine transformation that maps points1 onto points2.

    Arguments are specified as arrays of coordinates, shape (N+1, N).
    """
    N = points1.shape[1]
    N1 = N + 1
    if points2.shape != (N1, N) or points1.shape != (N1, N):
        raise TypeError("Points must have shape (N+1, N)")
    A = np.ones((N1, N1))
    A[:, :N] = points1
    B = np.ones((N1, N1))
    B[:, :N] = points2

    # solve N sets of linear equations to determine
    # transformation matrix elements
    matrix = np.eye(N1)
    for i in range(N):
        # solve Ax = B; x is one row of the desired transformation matrix
        matrix[i] = np.linalg.solve(A, B[:, i])

    return matrix[:N]


def bilinear_map2d(points1, points2):
    """
    Find a bilinear transformation matrix (shape 2x4) that maps 2D points1 onto 2D points2.

    Parameters
    ----------
    points1 :  array-like (N, 2)
        Array of N+1 2D points
    points2 :  array-like (N, 2)
        Array of N+1 2D points
    
    To use this matrix to map a point [x,y]::
    
        mapped = np.dot(matrix, [x*y, x, y, 1])
    """
    ## A is 4 rows (points) x 4 columns (xy, x, y, 1)
    ## B is 4 rows (points) x 2 columns (x, y)
    A = np.array([[points1[i,0]*points1[i,1], points1[i,0], points1[i,1], 1] for i in range(4)])
    B = np.array([[points2[i,0], points2[i,1]] for i in range(4)])
    
    ## solve 2 sets of linear equations to determine transformation matrix elements
    matrix = np.zeros((2,4))
    for i in range(2):
        matrix[i] = numpy.linalg.solve(A, B[:,i])  ## solve Ax = B; x is one row of the desired transformation matrix
    
    return matrix


def frustum(left, right, bottom, top, znear, zfar):
    """Create view frustum

    Parameters
    ----------
    left : float
        Left coordinate of the field of view.
    right : float
        Right coordinate of the field of view.
    bottom : float
        Bottom coordinate of the field of view.
    top : float
        Top coordinate of the field of view.
    znear : float
        Near coordinate of the field of view.
    zfar : float
        Far coordinate of the field of view.

    Returns
    -------
    M : ndarray
        View frustum matrix (4x4).
    """
    assert(right != left)
    assert(bottom != top)
    assert(znear != zfar)

    M = np.zeros((4, 4))
    M[0, 0] = +2.0 * znear / float(right - left)
    M[2, 0] = (right + left) / float(right - left)
    M[1, 1] = +2.0 * znear / float(top - bottom)
    M[2, 1] = (top + bottom) / float(top - bottom)
    M[2, 2] = -(zfar + znear) / float(zfar - znear)
    M[3, 2] = -2.0 * znear * zfar / float(zfar - znear)
    M[2, 3] = -1.0
    return M



from ._types import Mappable

from .base_transform import Transform, DependentTransformError
from .params import TransformListParameter


class CompositeTransform(Transform):
    """
    Transform subclass that performs a sequence of transformations in
    order.

    Arguments:

    transforms : list of Transform instances
        See ``transforms`` property.
    """

    Linear = False
    Orthogonal = False
    NonScaling = False
    Isometric = False
    Equidimensional = False

    parameter_spec = [
        TransformListParameter("transforms"),
    ]

    def __init__(self, transforms=None, **kwargs):
        super().__init__(transforms=transforms, **kwargs)

    def _init_with_no_state(self):
        super()._init_with_no_state()
        self._simplified = None

    @property
    def transforms(self):
        """The list of transform that make up the transform chain.

        The order of transforms is given such that the first transform in the
        list is the first to be invoked when mapping coordinates through
        the chain.

        For example, the following two mappings are equivalent::

            # Map coordinates through individual transforms:
            trans1 = STTransform(scale=(2, 3), translate=(0, 1))
            trans2 = PolarTransform()
            mapped = trans2.map(trans1.map(coords))

            # Equivalent mapping through chain:
            chain = CompositeTransform([trans1, trans2])
            mapped = chain.map(coords)

        """
        return self._state['transforms']

    @transforms.setter
    def transforms(self, value):
        self.set_params(transforms=value)

    def _validate_dims(self, dims, **kwargs):
        return None

    @property
    def systems(self):
        return self.transforms[0].systems[0], self.transforms[-1].systems[1]

    def set_systems(self, from_cs, to_cs, cs_graph=None):
        raise DependentTransformError("Cannot set systems on a CompositeTransform")

    def copy(self, from_cs=None, to_cs=None):
        if from_cs is not None or to_cs is not None:
            raise ValueError("Cannot set systems on a CompositeTransform")
        return super().copy()

    @property
    def dims(self):
        if len(self.transforms) == 0:
            return (None, None)
        return self.transforms[-1].dims[0], self.transforms[0].dims[1]

    def set_params(self, transforms=None):
        self._state.setdefault("transforms", [])
        from . import create_transform

        transforms = [t if isinstance(t, Transform) else create_transform(**t) for t in transforms or []]
        for i in range(len(transforms) - 1):
            if transforms[i].systems[1] != transforms[i + 1].systems[0]:
                raise TypeError(
                    f"Coordinate systems of transform {transforms[i]} "
                    f"({transforms[i].systems[1]}) does not map to {transforms[i+1]} "
                    f"({transforms[i+1].systems[0]})"
                )

        # Avoid extra effort if we already have the correct chain
        curr_xforms = self._state["transforms"]
        if len(transforms) == len(curr_xforms):
            changed = any(transforms[i] is not curr_xforms[i] for i in range(len(transforms)))
            if not changed:
                return

        for t in curr_xforms:
            t.remove_change_callback(self._subtr_changed)
        self._state["transforms"] = transforms
        self._simplified = None
        for t in transforms:
            t.add_change_callback(self._subtr_changed, keep_reference=False, duplicates='ignore')
        self._update()

    @property
    def simplified(self):
        """A simplified representation of the same transformation."""
        if self._simplified is None:
            self._simplified = SimplifiedCompositeTransform(self)
        return self._simplified

    @property
    def Linear(self):
        b = True
        for tr in self.transforms:
            b &= tr.Linear
        return b

    @property
    def Orthogonal(self):
        b = True
        for tr in self.transforms:
            b &= tr.Orthogonal
        return b

    @property
    def NonScaling(self):
        b = True
        for tr in self.transforms:
            b &= tr.NonScaling
        return b

    @property
    def Isometric(self):
        b = True
        for tr in self.transforms:
            b &= tr.Isometric
        return b

    def map(self, obj: Mappable):
        return self._map(obj)

    def _map(self, coords):
        """Map coordinates

        Parameters
        ----------
        coords : array-like
            Coordinates to map.

        Returns
        -------
        coords : ndarray
            Coordinates.
        """
        for tr in self.transforms:
            coords = tr.map(coords)
        return coords

    def _imap(self, coords):
        """Inverse map coordinates

        Parameters
        ----------
        coords : array-like
            Coordinates to inverse map.

        Returns
        -------
        coords : ndarray
            Coordinates.
        """
        for tr in reversed(self.transforms):
            coords = tr.imap(coords)
        return coords

    def as_affine(self):
        ret = None
        for tr in self.transforms:
            if ret is None:
                ret = tr.as_affine()
            else:
                ret = tr.as_affine() * ret
        return ret

    @property
    def full_matrix(self):
        mat = None
        for tr in self.transforms:
            if mat is None:
                mat = tr.full_matrix
            else:
                mat = tr.full_matrix.dot(mat)
        return mat

    def as_vispy(self):
        from vispy.visuals.transforms import ChainTransform

        return ChainTransform([tr.as_vispy() for tr in reversed(self.transforms)])

    def append(self, tr):
        """
        Add a new transform to the end of this chain.

        Parameters
        ----------
        tr : instance of Transform
            The transform to use.
        """
        print("appending", tr)
        self.transforms.append(tr)
        print("appended?", self.transforms)
        tr.add_change_callback(self._subtr_changed, keep_reference=False, duplicates='ignore')
        self._update()

    def prepend(self, tr):
        """
        Add a new transform to the beginning of this chain.

        Parameters
        ----------
        tr : instance of Transform
            The transform to use.
        """
        self.transforms.insert(0, tr)
        tr.add_change_callback(self._subtr_changed, keep_reference=False, duplicates='ignore')
        self._update()

    def _subtr_changed(self, event):
        """One of the internal transforms changed; propagate the signal."""
        self._update(event)

    def __setitem__(self, index, tr):
        self.transforms[index].remove_change_callback(self._subtr_changed)
        self.transforms[index] = tr
        tr.add_change_callback(self._subtr_changed, keep_reference=False, duplicates='ignore')
        self._update()

    def __mul__(self, tr):
        if isinstance(tr, CompositeTransform):
            trs = tr.transforms
        else:
            trs = [tr]
        return CompositeTransform(trs + self.transforms)

    def __rmul__(self, tr):
        if isinstance(tr, CompositeTransform):
            trs = tr.transforms
        else:
            trs = [tr]
        return CompositeTransform(self.transforms + trs)

    def __eq__(self, b):
        if not isinstance(b, CompositeTransform):
            return False
        return all(t1 == t2 for t1, t2 in zip(self.transforms, b.transforms))

    def __str__(self):
        names = [tr.__class__.__name__ for tr in self.transforms]
        return f"<{self.__class__.__name__} [{', '.join(names)}] at 0x{id(self):x}>"

    def __repr__(self):
        tr = ",\n                 ".join(map(repr, self.transforms))
        return f"<{self.__class__.__name__} [{tr}] at 0x{id(self):x}>"


class SimplifiedCompositeTransform(CompositeTransform):
    def __init__(self, chain):
        CompositeTransform.__init__(self)
        self._chain = chain
        chain.add_change_callback(self.source_changed, keep_reference=False, duplicates='ignore')
        self.source_changed(None)

    def source_changed(self, event):
        """Generate a simplified chain by joining adjacent transforms."""
        # bail out early if the chain is empty
        transforms = self._chain.transforms[:]
        if len(transforms) == 0:
            self.transforms = []
            return

        # If the change signal comes from a transform that already appears in
        # our simplified transform list, then there is no need to re-simplify.
        if event is not None:
            for source in event.sources[::-1]:
                if source in self.transforms:
                    self._update(event)
                    return

        # First flatten the chain by expanding all nested chains
        new_chain = []
        while len(transforms) > 0:
            tr = transforms.pop(0)
            if isinstance(tr, CompositeTransform) and not tr.dynamic:
                transforms = tr.transforms[:] + transforms
            else:
                new_chain.append(tr)

        # Now combine together all compatible adjacent transforms
        cont = True
        tr = new_chain
        while cont:
            new_tr = [tr[0]]
            cont = False
            for t2 in tr[1:]:
                t1 = new_tr[-1]
                pr = t2 * t1
                if not t1.dynamic and not t2.dynamic and not isinstance(pr, CompositeTransform):
                    cont = True
                    new_tr.pop()
                    new_tr.append(pr)
                else:
                    new_tr.append(t2)
            tr = new_tr

        self.transforms = tr



import os
import re

from setuptools import setup, find_packages

packages = [x for x in find_packages(".") if x.startswith("coorx")]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def get_version():
    init_py = os.path.join(os.path.dirname(__file__), "coorx", "__init__.py")
    with open(init_py, "r") as f:
        content = f.read()
    if version_match := re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", content, re.M):
        return version_match[1]
    raise RuntimeError("Unable to find version string")


setup(
    name="coorx",
    version=get_version(),
    author="Luke Campagnola",
    author_email="lukec@alleninstitute.org",
    description="Object-oriented linear and nonlinear coordinate system transforms, plus coordinate system graphs.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="BSD",
    url="http://github.com/campagnola/coorx",
    packages=packages,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
)
